/*******************************************************************************
**3456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789
**      10        20        30        40        50        60        70        80
**
** project:
**    OpenGL/cairo anti-aliasing trick
**
** file:
**    cairo-utils.c
**
** author:
**    Mirco "MacSlow" Mueller <macslow@bangang.de>
**
** copyright (C) Mirco Mueller, December 2006, placed under the terms of the GPL
**
*******************************************************************************/

#include <string.h>
#include <math.h>
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>

#include "cairo-utils.h"

/* just draw a path resembling a rectangle with rounded corners; there are no
** checks on width, height and corner-radius performed */
void
cairo_draw_round_rect (cairo_t* pContext,
		       double fAspectRatio,  /* aspect-ratio            */
		       double fX,            /* top-left corner         */
		       double fY,            /* top-left corner         */
		       double fCornerRadius, /* "size" of the corners   */
		       double fWidth,        /* width of the rectangle  */
		       double fHeight        /* height of the rectangle */)
{
	double fRadius = fCornerRadius / fAspectRatio;

	/* top-left, right of the corner */
	cairo_move_to (pContext, fX + fRadius, fY);

	/* top-right, left of the corner */
	cairo_line_to (pContext,
		       fX + fWidth - fRadius,
		       fY);

	/* top-right, below the corner */
	cairo_arc (pContext,
                   fX + fWidth - fRadius,
                   fY + fRadius,
                   fRadius,
                   -90.0f * G_PI / 180.0f,
                   0.0f * G_PI / 180.0f);

	/* bottom-right, above the corner */
	cairo_line_to (pContext,
		       fX + fWidth,
		       fY + fHeight - fRadius);

	/* bottom-right, left of the corner */
	cairo_arc (pContext,
                   fX + fWidth - fRadius,
                   fY + fHeight - fRadius,
                   fRadius,
                   0.0f * G_PI / 180.0f,
                   90.0f * G_PI / 180.0f);

	/* bottom-left, right of the corner */
	cairo_line_to (pContext,
		       fX + fRadius,
		       fY + fHeight);

	/* bottom-left, above the corner */
	cairo_arc (pContext,
                   fX + fRadius,
                   fY + fHeight - fRadius,
                   fRadius,
                   90.0f * G_PI / 180.0f,
                   180.0f * G_PI / 180.0f);

	/* top-left, below the corner */
	cairo_line_to (pContext,
		       fX,
		       fY + fRadius);

	/* top-left, right of the corner */
	cairo_arc (pContext,
                   fX + fRadius,
                   fY + fRadius,
                   fRadius,
                   180.0f * G_PI / 180.0f,
                   270.0f * G_PI / 180.0f);
}

gint
to_int (gdouble fValue)
{
	fValue = floor (fValue * 256);

	if (fValue > 255)
		return 255;
	else
		return (gint) fValue;
}

gdouble
d1 (gint iValue)
{
	return (gdouble) iValue / 255.0f;
}

gdouble
d2 (gint iValue)
{
	return ((gdouble) iValue + 0.5f) / 256.0f;
}

guchar
clamp (gdouble fValue)
{
	guchar ucValue;

	if (fValue > 255.0f)
		ucValue = 255;
	else if (fValue < 0.0f)
		ucValue = 0;
	else
		ucValue = (guchar) fValue;

	return ucValue;
}

void
compute_1d_kernel (gint	   iRadius,    /* radius of kernel                  */
		   gfloat  fDeviation, /* deviation to use, 0.0f for auto   */
		   gfloat* pfKernel    /* pointer to array for coefficients */)
{
	gfloat	 fSum     = 0.0f;
	gfloat	 fValue   = 0.0f;
	gint	 i;
	gint	 iSize	 = 2 * iRadius + 1;
	gfloat	 fRadius;

	/* sanity check */
	g_assert (pfKernel != NULL);

	if (iRadius <= 0)
		return;

	fRadius = fabs (iRadius) + 1.0f;
	if (fDeviation == 0.0f)
		fDeviation = sqrt (-(fRadius * fRadius) / (2.0f * log (1.0f / 255.0f)));

	fValue = (gfloat) -iRadius;
	for (i = 0; i < iSize; i++)
	{
		pfKernel[i] = 1.0f / (2.506628275f * fDeviation) *
				  expf (-((fValue * fValue) /
				  	  (2.0f * fDeviation * fDeviation)));
		fSum += pfKernel[i];
		fValue += 1.0f;
	}

	for (i = 0; i < iSize; i++)
		pfKernel[i] /= fSum;

	/*g_print ("std. deviation: %f\n", fDeviation);
	g_print ("kernel sum: %f\n\n", fSum);
	for (i = 0; i < iSize; i++)
		g_print ("%f - ", pfKernel[i]);
	g_print ("\n\n");*/
}

gdouble*
kernel_1d_new (gint    iRadius,	  /* radius of kernel                                */
	       gdouble fDeviation /* deviation to use, pass 0.0f for auto-generation */)
{
	gdouble* pfKernel = NULL;
	gdouble	 fSum     = 0.0f;
	gdouble	 fValue   = 0.0f;
	gint	 i;
	gint	 iSize	 = 2 * iRadius + 1;
	gdouble	 fRadius;

	if (iRadius <= 0)
		return NULL;

	pfKernel = (gdouble*) g_malloc0 ((iSize + 1) * sizeof (gdouble));
	if (!pfKernel)
		return NULL;

	fRadius = fabs (iRadius) + 1.0f;
	if (fDeviation == 0.0f)
		fDeviation = sqrt (-(fRadius * fRadius) / (2.0f * log (1.0f / 255.0f)));

	/*g_print ("std. deviation: %f\n", fDeviation);*/

	pfKernel[0] = iSize;
	fValue = (gdouble) -iRadius;
	for (i = 0; i < iSize; i++)
	{
		pfKernel[1 + i] = 1.0f / (2.506628275f * fDeviation) *
				  expf (-((fValue * fValue) /
				  	  (2.0f * fDeviation * fDeviation)));
		fSum += pfKernel[1 + i];
		fValue += 1.0f;
	}

	for (i = 0; i < iSize; i++)
		pfKernel[1 + i] /= fSum;

	/*g_print ("kernel sum: %f\n\n", fSum);

	for (i = 0; i < iSize; i++)
		g_print ("%f - ", pfKernel[1 + i]);

	g_print ("\n\n");*/

	return pfKernel;
}

void
kernel_1d_delete (gdouble* pfKernel)
{
	/* sanity check */
	g_assert (pfKernel != NULL);

	g_free ((gpointer) pfKernel);
}

void
convolute (guchar*  pucImageBuffer, /* pointer to image-data */
	   gint	    iImageWidth,    /* width of image        */
	   gint	    iImageHeight,   /* height of image       */
	   gint	    iImageStride,   /* stride of image       */
	   gdouble* pfKernel        /* the kernel            */)
{
	guchar*	pucResultBuffer = NULL;
	gint	iX;
	gint	iY;
	gint	iKernelOffsetX;
	gint	iKernelOffsetY;
	gint	iKernelOffset;
	gint	iPixelOffset;
	gint	iHalfWidth = ((gint) pfKernel[0] - 1) / 2;
	gint	iHalfHeight = ((gint) pfKernel[1] - 1) / 2;
	gdouble	fSum = 0.0f;
	gdouble	fBias = 0.0f;

	/* sanity check */
	g_assert (pucImageBuffer != NULL);
	g_assert (pfKernel != NULL);

	for (iKernelOffsetY = 0;
	     iKernelOffsetY < (gint) pfKernel[1];
	     iKernelOffsetY++)
	{
		for (iKernelOffsetX = 0;
		     iKernelOffsetX < (gint) pfKernel[0];
		     iKernelOffsetX++)
		{
			iKernelOffset = 3 +
					iKernelOffsetY *
					(gint) pfKernel[0]+
					iKernelOffsetX;
			fSum += pfKernel[iKernelOffset];
		}
	}

	if (fSum == 0.0f)
		fBias = 128.0f;

	/* create buffer to hold the convoluted image */
	pucResultBuffer = (guchar*) g_malloc0 (iImageHeight * iImageStride);
	if (!pucResultBuffer)
	{
		g_print ("convolute(): ");
		g_print ("Could not allocate the temporary image buffer!\n");
		return;
	}

	for (iY = 0; iY < iImageHeight; iY++)
	{
		for (iX = 0; iX < iImageWidth; iX++)
		{
			gfloat fRed   = 0.0f;
			gfloat fGreen = 0.0f;
			gfloat fBlue  = 0.0f;
			gfloat fAlpha = 0.0f;

			for (iKernelOffsetY = 0;
			     iKernelOffsetY < (gint) pfKernel[1];
			     iKernelOffsetY++)
			{
				for (iKernelOffsetX = 0;
				     iKernelOffsetX < (gint) pfKernel[0];
				     iKernelOffsetX++)
				{
					if ((iKernelOffsetX - iHalfWidth + iX) >= 0 &&
					    (iKernelOffsetX - iHalfWidth + iX) <= iImageWidth &&
					    (iKernelOffsetY - iHalfHeight + iY) >= 0 &&
					    (iKernelOffsetY - iHalfHeight + iY) <+ iImageHeight)
					{
						iKernelOffset = 3 + iKernelOffsetY * (gint) pfKernel[0] + iKernelOffsetX;
						if (pfKernel[iKernelOffset] == 0.0f)
							continue;

						iPixelOffset = (iY + iKernelOffsetY - iHalfHeight) * iImageStride + (iX + iKernelOffsetX - iHalfWidth) * 4;
						fAlpha += (pfKernel[iKernelOffset] / pfKernel[2] * (gdouble) pucImageBuffer[iPixelOffset + 3]);
						fRed   += (pfKernel[iKernelOffset] / pfKernel[2] * (gdouble) pucImageBuffer[iPixelOffset + 2]);
						fGreen += (pfKernel[iKernelOffset] / pfKernel[2] * (gdouble) pucImageBuffer[iPixelOffset + 1]);
						fBlue  += (pfKernel[iKernelOffset] / pfKernel[2] * (gdouble) pucImageBuffer[iPixelOffset + 0]);
					}
				}
			}

			pucResultBuffer[iY * iImageStride + iX * 4 + 3] = 255;
			pucResultBuffer[iY * iImageStride + iX * 4 + 2] = clamp (fRed);
			pucResultBuffer[iY * iImageStride + iX * 4 + 1] = clamp (fGreen);
			pucResultBuffer[iY * iImageStride + iX * 4 + 0] = clamp (fBlue);
		}
	}

	g_memmove (pucImageBuffer, pucResultBuffer, iImageHeight * iImageStride);
	g_free ((gpointer) pucResultBuffer);
}

void
glow (guchar* pucSrcBuffer, /* pointer to source             */
      gint    iSrcWidth,    /* width of source               */
      gint    iSrcHeight,   /* height of source              */
      gint    iSrcChannels, /* color-channels in source      */
      guchar* pucDstBuffer, /* pointer to destination        */
      gint*   piDstWidth,   /* pointer to destination width  */
      gint*   piDstHeight   /* pointer to destination height */)
{
	gint iSrcStride;
	gint iDstStride;
	gint iRadius;
	gint iX;
	gint iY;
	gint iSrcOffset;
	gint iDstOffset;
	cairo_surface_t* pSurface;
	/*gdouble afBlur[] = {5.0f, 5.0f, 100.0f, 1.0f, 2.0f,  4.0f, 2.0f, 1.0f,
						  2.0f, 4.0f,  8.0f, 4.0f, 2.0f,
						  4.0f, 8.0f, 16.0f, 8.0f, 4.0f,
						  2.0f, 4.0f,  8.0f, 4.0f, 2.0f,
						  1.0f, 2.0f,  4.0f, 2.0f, 1.0f};
	gdouble afSobelA[] = {3.0f, 3.0f, 1.0f, -1.0f, 0.0f, 1.0f,
						-2.0f, 0.0f, 2.0f,
						-1.0f, 0.0f, 1.0f};
	gdouble afSobelB[] = {3.0f, 3.0f, 1.0f,  1.0f,  2.0f,  1.0f,
						 0.0f,  0.0f,  0.0f,
						-1.0f, -2.0f, -1.0f};*/
	gdouble afSobelC[] = {3.0f, 3.0f, 1.0f,  0.0f,   1.25f, 1.25f,
						-1.25f,  0.0f,  1.25f,
						-1.25f, -1.25f, 0.0f};

	/* sanity check */
	g_assert (pucSrcBuffer != NULL);

	iRadius = 3;
	*piDstWidth = iSrcWidth + 2 * iRadius;
	*piDstHeight = iSrcHeight + 2 * iRadius;
	iDstStride = 4 * *piDstWidth;
	iSrcStride = iSrcChannels * iSrcWidth;

	/* create destination-buffer */
	pucDstBuffer = (guchar*) g_malloc0 (4 * *piDstWidth * *piDstHeight);
	if (!pucDstBuffer)
	{
		g_print ("glow(): ");
		g_print ("Could not allocate the glow buffer!\n");
		return;
	}

	/* copy source-image into destination-buffer */
	for (iY = 0; iY < iSrcHeight; iY++)
	{
		for (iX = 0; iX < iSrcWidth; iX++)
		{
			iSrcOffset = iY * iSrcStride + iX * iSrcChannels;
			iDstOffset = (iY + iRadius) * iDstStride + (iX + iRadius) * 4;

			pucDstBuffer[iDstOffset + 0] = pucSrcBuffer[iSrcOffset + 0];
			pucDstBuffer[iDstOffset + 1] = pucSrcBuffer[iSrcOffset + 1];
			pucDstBuffer[iDstOffset + 2] = pucSrcBuffer[iSrcOffset + 2];

			if (iSrcChannels == 4)
				pucDstBuffer[iDstOffset + 3] = pucSrcBuffer[iSrcOffset + 3];
			else
				pucDstBuffer[iDstOffset + 3] = 255;
		}
	}

	pSurface = cairo_image_surface_create_for_data (pucDstBuffer,
                                                        CAIRO_FORMAT_ARGB32,
                                                        *piDstWidth,
                                                        *piDstHeight,
                                                        iDstStride);
	cairo_surface_write_to_png (pSurface, "./before-glow.png");
	cairo_surface_destroy (pSurface);

	/* blur the destination-buffer */
	convolute (pucDstBuffer,
		   *piDstWidth,
		   *piDstHeight,
		   iDstStride,
		   afSobelC);
	/*blur (pucDstBuffer, *piDstWidth, *piDstHeight, 4, iRadius);*/

	pSurface = cairo_image_surface_create_for_data (pucDstBuffer,
                                                        CAIRO_FORMAT_ARGB32,
                                                        *piDstWidth,
                                                        *piDstHeight,
                                                        iDstStride);
	cairo_surface_write_to_png (pSurface, "./afer-glow.png");
	cairo_surface_destroy (pSurface);
}

void
blur (guchar* pucSrc,    /* pointer to image-data   */
      gint    iWidth,    /* width of image          */
      gint    iHeight,   /* height of image         */
      gint    iChannels, /* color-channels in image */
      gint    iRadius    /* the blur-radius         */)
{
	gint	 iX	       = 0;
	gint	 iY	       = 0;
	gint	 i	       = 0;
	gint	 x	       = 0;
	gint	 y	       = 0;
	gint	 iStride       = iWidth * iChannels;
	gint	 iOffset       = 0;
	gint	 iBaseOffset   = 0;
	gdouble* pfHorzBlurred = NULL;
	gdouble* pfVertBlurred = NULL;
	gdouble* pfKernel      = NULL;

	/* sanity check */
	g_assert (pucSrc != NULL);

	/* only process 3- or 4-channel image */
	if (!(iChannels == 3 || iChannels == 4))
		return;

	/* create buffers to hold the blur-passes */
	pfHorzBlurred = (gdouble*) g_malloc0 (iHeight *
					      iStride *
					      sizeof (gdouble));
	pfVertBlurred = (gdouble*) g_malloc0 (iHeight *
					      iStride *
					      sizeof (gdouble));
	if (!pfHorzBlurred || !pfVertBlurred)
	{
		g_print ("blur(): ");
		g_print ("Could not allocate the temporary blur-buffers!\n");
		return;
	}

	pfKernel = kernel_1d_new (iRadius, 0.0f);

	/* horizontal pass */
	for (iY = 0; iY < iHeight; iY++)
	{
		for (iX = 0; iX < iWidth; iX++)
		{
			gdouble fRed   = 0.0f;
			gdouble fGreen = 0.0f;
			gdouble fBlue  = 0.0f;
			gdouble fAlpha = 0.0f;

			iOffset = ((gint) pfKernel[0]) / -2;
			for (i = 0; i < (gint) pfKernel[0]; i++)
			{
				x = iX + iOffset;
				if (x >= 0 && x <= iWidth)
				{
					iBaseOffset = iY * iStride +
						      x * iChannels;

					if (iChannels == 4)
						fAlpha += (pfKernel[1+i] *
							   (gdouble) pucSrc[iBaseOffset + 3]);

					fRed   += (pfKernel[1+i] *
						   (gdouble) pucSrc[iBaseOffset + 2]);

					fGreen += (pfKernel[1+i] *
						   (gdouble) pucSrc[iBaseOffset + 1]);

					fBlue  += (pfKernel[1+i] *
						   (gdouble) pucSrc[iBaseOffset + 0]);
				}

				iOffset++;
			}

			iBaseOffset = iY * iStride + iX * iChannels;

			if (iChannels == 4)
				pfHorzBlurred[iBaseOffset + 3] = fAlpha;

			pfHorzBlurred[iBaseOffset + 2] = fRed;
			pfHorzBlurred[iBaseOffset + 1] = fGreen;
			pfHorzBlurred[iBaseOffset + 0] = fBlue;
		}
	}

	/* vertical pass */
	for (iY = 0; iY < iHeight; iY++)
	{
		for (iX = 0; iX < iWidth; iX++)
		{
			gdouble fRed   = 0.0f;
			gdouble fGreen = 0.0f;
			gdouble fBlue  = 0.0f;
			gdouble fAlpha = 0.0f;

			iOffset = ((gint) pfKernel[0]) / -2;
			for (i = 0; i < (gint) pfKernel[0]; i++)
			{
				y = iY + iOffset;
				if (y >= 0 && y <= iHeight)
				{
					iBaseOffset = y * iStride +
						      iX * iChannels;

					if (iChannels == 4)
						fAlpha += (pfKernel[1+i] *
							   pfHorzBlurred[iBaseOffset + 3]);

					fRed   += (pfKernel[1+i]  *
						   pfHorzBlurred[iBaseOffset + 2]);

					fGreen += (pfKernel[1+i]  *
						   pfHorzBlurred[iBaseOffset + 1]);

					fBlue  += (pfKernel[1+i]  *
						   pfHorzBlurred[iBaseOffset + 0]);
				}

				iOffset++;
			}

			iBaseOffset = iY * iStride + iX * iChannels;

			if (iChannels == 4)
				pfVertBlurred[iBaseOffset + 3] = fAlpha;

			pfVertBlurred[iBaseOffset + 2] = fRed;
			pfVertBlurred[iBaseOffset + 1] = fGreen;
			pfVertBlurred[iBaseOffset + 0] = fBlue;
		}
	}

	kernel_1d_delete (pfKernel);

	for (iY = 0; iY < iHeight; iY++)
	{
		for (iX = 0; iX < iWidth; iX++)
		{
			iOffset = iY * iStride + iX * iChannels;

			if (iChannels == 4)
				pucSrc[iOffset + 3] = (guchar) pfVertBlurred[iOffset + 3];

			pucSrc[iOffset + 2] = (guchar) pfVertBlurred[iOffset + 2];
			pucSrc[iOffset + 1] = (guchar) pfVertBlurred[iOffset + 1];
			pucSrc[iOffset + 0] = (guchar) pfVertBlurred[iOffset + 0];
		}
	}

	g_free ((gpointer) pfHorzBlurred);
	g_free ((gpointer) pfVertBlurred);
}

gdouble*
build_image_sat (guchar* pucSrcBuffer,
		 gint	 iWidth,
		 gint	 iHeight,
		 gint	 iChannels)
{
	gint	 iX	     = 0;
	gint	 iY	     = 0;
	gint	 iStride     = iWidth * iChannels;
	gint	 iPixel	     = 0;
	gdouble* pfSATBuffer = NULL;
	gdouble  afCurrent[4];
	gdouble  afLast[4];

	/* sanity check */
	g_assert (pucSrcBuffer != NULL);

	/* only process 3- or 4-channel images */
	if (!(iChannels == 3 || iChannels == 4))
		return NULL;

	/* create buffer to hold the blurred image */
	pfSATBuffer = (gdouble*) g_malloc0 (iHeight * iStride * sizeof (gdouble));
	if (!pfSATBuffer)
	{
		g_print ("build_image_sat(): ");
		g_print ("Could not allocate the temporary buffer!\n");
		return NULL;
	}

	/* first pass, sum up rows */
	for (iY = 0; iY < iHeight; iY++)
	{
		afLast[R] = 0.0f;
		afLast[G] = 0.0f;
		afLast[B] = 0.0f;
		afLast[A] = 0.0f;

		for (iX = 0; iX < iWidth; iX++)
		{
			iPixel = iY * iStride + iX * iChannels;

			afCurrent[R] = (gdouble) pucSrcBuffer[iPixel + R] / 255.0f;
			afCurrent[G] = (gdouble) pucSrcBuffer[iPixel + G] / 255.0f;
			afCurrent[B] = (gdouble) pucSrcBuffer[iPixel + B] / 255.0f;
			afCurrent[A] = (gdouble) pucSrcBuffer[iPixel + A] / 255.0f;

			pfSATBuffer[iPixel + R] = afCurrent[R] + afLast[R];
			pfSATBuffer[iPixel + G] = afCurrent[G] + afLast[G];
			pfSATBuffer[iPixel + B] = afCurrent[B] + afLast[B];
			pfSATBuffer[iPixel + A] = afCurrent[A] + afLast[A];

			afLast[R] = pfSATBuffer[iPixel + R];
			afLast[G] = pfSATBuffer[iPixel + G];
			afLast[B] = pfSATBuffer[iPixel + B];
			afLast[A] = pfSATBuffer[iPixel + A];
		}
	}

	/* second pass, sum up columns */
	for (iX = 0; iX < iWidth; iX++)
	{
		afLast[R] = 0.0f;
		afLast[G] = 0.0f;
		afLast[B] = 0.0f;
		afLast[A] = 0.0f;

		for (iY = 0; iY < iHeight; iY++)
		{
			iPixel = iY * iStride + iX * iChannels;

			afCurrent[R] = pfSATBuffer[iPixel + R];
			afCurrent[G] = pfSATBuffer[iPixel + G];
			afCurrent[B] = pfSATBuffer[iPixel + B];
			afCurrent[A] = pfSATBuffer[iPixel + A];

			pfSATBuffer[iPixel + R] = afCurrent[R] + afLast[R];
			pfSATBuffer[iPixel + G] = afCurrent[G] + afLast[G];
			pfSATBuffer[iPixel + B] = afCurrent[B] + afLast[B];
			pfSATBuffer[iPixel + A] = afCurrent[A] + afLast[A];

			afLast[R] = pfSATBuffer[iPixel + R];
			afLast[G] = pfSATBuffer[iPixel + G];
			afLast[B] = pfSATBuffer[iPixel + B];
			afLast[A] = pfSATBuffer[iPixel + A];
		}
	}

	return pfSATBuffer;
}

void
demultiply_alpha (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */)
{
	gint iX = 0;
	gint iY = 0;

	/* sanity check */
	g_assert (pucPixelBuffer != NULL);

	/* get rid of premultiplied-alpha for use in OpenGL */
	for (iY = 0; iY < iHeight; iY++)
	{
		for (iX = 0; iX < iWidth; iX++)
		{
			guchar ucRed;
			guchar ucGreen;
			guchar ucBlue;
			guchar ucAlpha;

			ucAlpha = pucPixelBuffer[iY * iStride + iX * 4 + 3];
			if (ucAlpha == 0 || ucAlpha == 255)
				continue;

			ucRed = pucPixelBuffer[iY * iStride + iX * 4 + 2];
			ucGreen = pucPixelBuffer[iY * iStride + iX * 4 + 1];
			ucBlue = pucPixelBuffer[iY * iStride + iX * 4 + 0];

			ucRed	= to_int (d2 (ucRed) / d1 (ucAlpha));
			ucGreen	= to_int (d2 (ucGreen) / d1 (ucAlpha));
			ucBlue	= to_int (d2 (ucBlue) / d1 (ucAlpha));

			pucPixelBuffer[iY * iStride + iX * 4 + 0] = ucBlue;
			pucPixelBuffer[iY * iStride + iX * 4 + 1] = ucGreen;
			pucPixelBuffer[iY * iStride + iX * 4 + 2] = ucRed;
		}
	}
}

/* one of my special tricks for smooth solid looking graphics, antialising
** without the actual need for aa-hardware, thus very fast and still looking
** perfect; the idea is to leverage the texture-filtering hardware of GL-cards
** to smooth out the edges of a specially prepared texture; slightly trim the
** edges of a texture using cairo's image-surfaces and while doing that also add
** some nicely rounded corners; thanks to cairo's antialised vector-drawing the
** edges are faded out in a subtle way, no hard jaggies; this should be applied
** _before_ the image-data is actually transfered to GL as texture; the supplied
** image-data has to be R8G8B8A8; it only works (or makes sense) for single
** quads/rects/etc */
void
cairo_trim_image (guchar* pucPixelBuffer, /* pointer to image-data */
		  gint iWidth,            /* width of image        */
		  gint iHeight,           /* height of image       */
		  gint iStride            /* stride of image       */)
{	
	cairo_surface_t* pSurface = NULL;
	cairo_status_t status = CAIRO_STATUS_SUCCESS;
	cairo_t* pContext = NULL;
	cairo_pattern_t* pPattern = NULL;

	/* sanity check */
	g_assert (pucPixelBuffer != NULL);

	/* create surface from supplied texture-data */
	pSurface = cairo_image_surface_create_for_data (pucPixelBuffer,
							CAIRO_FORMAT_ARGB32,
							iWidth,
							iHeight,
							iStride);

	/* check for errors */
	status = cairo_surface_status (pSurface);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* create the rendering context */
	pContext = cairo_create (pSurface);

	/* check for any errors */
	status = cairo_status (pContext);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pSurface);
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* ensure that we can use normalized coordinates */
	cairo_scale (pContext, (double) iWidth, (double) iHeight);

	/* create a pattern */
	pPattern = cairo_pattern_create_rgba  (0.0f, 0.0f, 0.0f, 1.0f);
	status = cairo_pattern_status (pPattern);
	if (status != CAIRO_STATUS_SUCCESS)
	{
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_print ("%s\n", cairo_status_to_string (status));
		return;
	}

	/* now use the pattern to draw a pattern-filled rounded rectangle */
	cairo_set_operator (pContext, CAIRO_OPERATOR_DEST_IN);
	cairo_set_source (pContext, pPattern);
	cairo_draw_round_rect (pContext,
			       1.0f,
			       0.01f,
			       0.01f,
			       0.03f,
			       0.98f,
			       0.98f);	
	cairo_fill (pContext);

	/* get rid of premultiplied-alpha for use in OpenGL */
	demultiply_alpha (pucPixelBuffer, iWidth, iHeight, iStride);

	/* clean up */
	cairo_pattern_destroy (pPattern);
	cairo_destroy (pContext);
	cairo_surface_destroy (pSurface);
}

Texture*
cairo_text_to_texture (gchar* pcText /* the text to render */)
{
	guchar*			pucBufferData = NULL;
	gint			iBufferSize = 0;
	cairo_surface_t*	pTempSurface = NULL;
	cairo_t*		pTempContext = NULL;
	PangoFontDescription*	pDesc = NULL;
	PangoLayout*		pLayout = NULL;
	cairo_surface_t*	pTextSurface = NULL;
	cairo_t*		pTextContext = NULL;
	PangoRectangle		ink = {0, 0, 0, 0};
	PangoRectangle		log = {0, 0, 0, 0};
	Texture*		pTexture = NULL;

	/* sanity check */
	g_assert (pcText != NULL);

	/* setup some buffer for creating a temporary context */
	iBufferSize = 4;
	pucBufferData = (guchar*) g_malloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate the temporary buffer!\n");
		return NULL;
	}

	/* create surface from supplied buffer and check for errors */
	pTempSurface = cairo_image_surface_create_for_data (pucBufferData,
							    CAIRO_FORMAT_ARGB32,
							    1,
							    1,
							    4);
	if (cairo_surface_status (pTempSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create temporary surface!\n");
		return NULL;
	}

	/* create a temporary rendering context and check for errors */
	pTempContext = cairo_create (pTempSurface);
	if (cairo_status (pTempContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pTempSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not temporary context!\n");
		return NULL;
	}

	/* setup a new pango-layout based on the temporary context */
	pLayout = pango_cairo_create_layout (pTempContext);
	if (!pLayout)
	{
		cairo_destroy (pTempContext);
		cairo_surface_destroy (pTempSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create pango-layout!\n");
		return NULL;
	}

	/* get a new pango-description */
	pDesc = pango_font_description_new ();
	if (!pDesc)
	{
		g_object_unref (pLayout);
		cairo_destroy (pTempContext);
		cairo_surface_destroy (pTempSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create pango-font-description!\n");
		return NULL;
	}

	/* set some desired font-attributes */
	pango_font_description_set_absolute_size (pDesc, 24 * PANGO_SCALE);
	pango_font_description_set_family_static (pDesc,
						  "DejaVu LGC Sans");
	pango_font_description_set_weight (pDesc, PANGO_WEIGHT_SEMIBOLD);
	pango_layout_set_font_description (pLayout, pDesc);
	pango_font_description_free (pDesc);

	/* set the text to render and... */
	pango_layout_set_text (pLayout, pcText, -1);

	/* ... get size of rendered string in pixels */
	pango_layout_get_pixel_extents (pLayout, &ink, &log);

	/* get rid of the temporary context and buffer, no longer needed */
	cairo_destroy (pTempContext);
	cairo_surface_destroy (pTempSurface);
	g_free ((gpointer) pucBufferData);

	/* get a buffer for creating the text-context */
	iBufferSize = log.width * log.height * 4;
	pucBufferData = (guchar*) g_malloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate the temporary buffer!\n");
		return NULL;
	}

	/* create a new surface for the text */
	pTextSurface = cairo_image_surface_create_for_data (pucBufferData,
							    CAIRO_FORMAT_ARGB32,
							    log.width,
							    log.height,
							    log.width * 4);
	if (cairo_surface_status (pTextSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_free ((gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create text-surface!\n");
		return NULL;
	}

	/* create a context from that new text surface */
	pTextContext = cairo_create (pTextSurface);
	if (cairo_status (pTextContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pTextSurface);
		g_free ((gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not create drawing-context for text!\n");
		return NULL;
	}

	/* create our own texture object */
	pTexture = (Texture*) g_malloc0 (sizeof (Texture));
	if (!pTexture)
	{
		cairo_destroy (pTextContext);
		cairo_surface_destroy (pTextSurface);
		g_free ((gpointer) pucBufferData);
		g_object_unref (pLayout);
		g_print ("cairo_text_to_texture(): ");
		g_print ("Could not allocate memory for the texture struct!\n");
		return NULL;
	}

	/* clear the newly created text-context */
	cairo_set_operator (pTextContext, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (pTextContext, 0.0f, 0.0f, 0.0f, 0.0f);
	cairo_paint (pTextContext);

	/* draw the text on the the text-context */
	cairo_set_source_rgba (pTextContext, 1.0f, 1.0f, 0.7f, 1.0f);
	cairo_move_to (pTextContext, 0.0f, 0.0f);
	pango_cairo_layout_path (pTextContext, pLayout);
	cairo_fill (pTextContext);

	/* fill the resulting texture-struct */
	pTexture->iTarget = opengl_what_texture_target (log.width, log.height);
	pTexture->fWidth = (float) log.width;
	pTexture->fHeight = (float) log.height;
	pTexture->iBufferSize = iBufferSize;
	pTexture->pucBufferData = pucBufferData; /* to be freed by the caller */

	/* get rid of premultiplied-alpha for use in OpenGL */
	demultiply_alpha (pucBufferData, log.width, log.height, log.width * 4);

	/* free all no longer needed resources */
	cairo_destroy (pTextContext);
	cairo_surface_destroy (pTextSurface);
	g_object_unref (pLayout);

	return pTexture; /* to be freed by the caller */
}

/* TODO: not done yet */
Texture*
cairo_logos_to_texture (gchar* pcSvgFileOne, /* full path first SVG-file  */
			gchar* pcSvgFileTwo, /* full path second SVG-file */
			gint iWidth,	     /* width of window           */
			gint iHeight	     /* height of window          */)
{
	Texture* pTexture = NULL;
	/*RsvgHandle* pSvgOne = NULL;
	RsvgHandle* pSvgTwo = NULL;
	RsvgDimensionData dimensionsOne;
	RsvgDimensionData dimensionsTwo;
	double fAspectOne = 1.0f;
	double fAspectTwo = 1.0f;
	cairo_surface_t* pSurface = NULL;
	cairo_t* pContext = NULL;
	GError* pError = NULL;
	gint iBufferSize = 0;
	guchar* pucBufferData = NULL;

	* sanity checks *
	g_assert (pcSvgFileOne != NULL);
	g_assert (pcSvgFileTwo != NULL);

	* setup some buffer for creating a temporary context *
	iBufferSize = iWidth * iHeight * 4;
	pucBufferData = (guchar*) g_malloc0 (iBufferSize);
	if (!pucBufferData)
	{
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not allocate the buffer!\n");
		return NULL;
	}

	* create surface from supplied buffer and check for errors *
	pSurface = cairo_image_surface_create_for_data (pucBufferData,
							CAIRO_FORMAT_ARGB32,
							iWidth,
							iHeight,
							iWidth * 4);
	if (cairo_surface_status (pSurface) != CAIRO_STATUS_SUCCESS)
	{
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not create surface!\n");
		return NULL;
	}

	* create a temporary rendering context and check for errors *
	pContext = cairo_create (pSurface);
	if (cairo_status (pContext) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy (pSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could not context!\n");
		return NULL;
	}

	rsvg_init ();

	* load first SVG-file *
	pSvgOne = rsvg_handle_new_from_file (pcSvgFileOne, &pError);
	if (!pSvgOne = NULL)
	{
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could load first SVG-file!\n");
		return NULL;
	}

	* load second SVG-file *
	pSvgTwo = rsvg_handle_new_from_file (pcSvgFileTwo, &pError);
	if (!pSvgTwo)
	{
		rsvg_handle_free (pSvgOne);
		cairo_destroy (pContext);
		cairo_surface_destroy (pSurface);
		g_free ((gpointer) pucBufferData);
		g_print ("cairo_logos_to_texture(): ");
		g_print ("Could load second SVG-file!\n");
		return NULL;
	}

	* get dimensions of the two SVG-files *
	rsvg_handle_get_dimensions (pSvgOne, &dimensionsOne);
	rsvg_handle_get_dimensions (pSvgTwo, &dimensionsTwo);

	* sanity check - to avoid division by zero later on *
	if (dimensionsOne.width == 0)
		dimensionsOne.width = 1;

	if (dimensionsOne.height == 0)
		dimensionsOne.height = 1;

	if (dimensionsTwo.width == 0)
		dimensionsTwo.width = 1;

	if (dimensionsTwo.height == 0)
		dimensionsTwo.height = 1;

	* get aspect-ratios of both SVG-files *
	fAspectOne = (double) dimensionsOne.width /
		     (double) dimensionsOne.height;

	fAspectTwo = (double) dimensionsTwo.width /
		     (double) dimensionsTwo.height;

	cairo_scale (pContext,
		     (double) iWidth / (double) dimensionsOne.width,
		     (double) iHeight / (double) dimensionsOne.height);

	rsvg_handle_render_cairo (pSvgOne, pContext);
	rsvg_handle_render_cairo (pSvgTwo, pContext);

	rsvg_handle_free (pSvgOne);
	rsvg_handle_free (pSvgTwo);


	rsvg_term ();*/

	return pTexture;
}
